@@ -2,23 +2,16 @@ import sys
 import json
 import asyncio
 import logging
-import argparse
 
-import zigpy_znp.types as t
-import zigpy_znp.commands as c
 from zigpy_znp.api import ZNP
 from zigpy_znp.config import CONFIG_SCHEMA
-from zigpy_znp.types.nvids import ExNvIds, NvSysIds, OsalNvIds
-from zigpy_znp.tools.common import setup_parser
+from zigpy_znp.types.nvids import ExNvIds, OsalNvIds
+from zigpy_znp.tools.common import ClosableFileType, setup_parser
 
 LOGGER = logging.getLogger(__name__)
 
 
-async def restore(radio_path, backup):
-    znp = ZNP(CONFIG_SCHEMA({"device": {"path": radio_path}}))
-
-    await znp.connect()
-
+async def nvram_write(znp: ZNP, backup):
     # First write the NVRAM items common to all radios
     for nwk_nvid, value in backup["LEGACY"].items():
         if "+" in nwk_nvid:
@@ -42,7 +35,6 @@ async def restore(radio_path, backup):
             value = bytes.fromhex(value)
 
             await znp.nvram.write(
-                sys_id=NvSysIds.ZSTACK,
                 item_id=item_id,
                 sub_id=sub_id,
                 value=value,
@@ -50,21 +42,24 @@ async def restore(radio_path, backup):
             )
 
     # Reset afterwards to have the new values take effect
-    await znp.request_callback_rsp(
-        request=c.SYS.ResetReq.Req(Type=t.ResetType.Soft),
-        callback=c.SYS.ResetInd.Callback(partial=True),
-    )
+    await znp.reset()
 
 
 async def main(argv):
     parser = setup_parser("Restore a radio's NVRAM from a previous backup")
     parser.add_argument(
-        "--input", "-i", type=argparse.FileType("r"), help="Input file", required=True
+        "--input", "-i", type=ClosableFileType("r"), help="Input file", required=True
     )
 
     args = parser.parse_args(argv)
-    backup = json.load(args.input)
-    await restore(args.serial, backup)
+
+    with args.input as f:
+        backup = json.load(f)
+
+    znp = ZNP(CONFIG_SCHEMA({"device": {"path": args.serial}}))
+    await znp.connect()
+
+    await nvram_write(znp, backup)
 
 
 if __name__ == "__main__":
