@@ -1,14 +1,30 @@
+from __future__ import annotations
+
 import sys
 import typing
 import logging
 import dataclasses
 
-from zigpy.types import NWK, EUI64, PanId, KeyData, ClusterId, ExtendedPanId
+from zigpy.types import (  # noqa: F401
+    NWK,
+    EUI64,
+    Bool,
+    PanId,
+    Struct,
+    KeyData,
+    Channels,
+    ClusterId,
+    ExtendedPanId,
+    CharacterString,
+)
+from zigpy.zdo.types import Status as ZDOStatus  # noqa: F401
 
-from . import basic, struct
+from . import basic
 
 LOGGER = logging.getLogger(__name__)
 
+JSONType = typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]
+
 
 class ADCChannel(basic.enum_uint8):
     """The ADC channel."""
@@ -68,51 +84,60 @@ class AddrMode(basic.enum_uint8):
     Broadcast = 0x0F
 
 
-class AddrModeAddress(struct.Struct):
-    mode: AddrMode
-    address: typing.Union[NWK, EUI64] = struct.StructField(
-        dynamic_type=lambda s: {
+class AddrModeAddress:
+    def __new__(cls, mode=None, address=None):
+        if mode is not None and address is None and isinstance(mode, cls):
+            other = mode
+            return cls(mode=other.mode, address=other.address)
+
+        instance = super().__new__(cls)
+
+        if mode is not None and mode == AddrMode.NOT_PRESENT:
+            raise ValueError(f"Invalid address mode: {mode}")
+
+        instance.mode = None if mode is None else AddrMode(mode)
+        instance.address = (
+            None if address is None else instance._get_address_type()(address)
+        )
+
+        return instance
+
+    def _get_address_type(self):
+        return {
             AddrMode.NWK: NWK,
             AddrMode.Group: NWK,
             AddrMode.Broadcast: NWK,
             AddrMode.IEEE: EUI64,
-        }[s.mode]
-    )
+        }[self.mode]
 
     @classmethod
     def deserialize(cls, data: bytes) -> "AddrModeAddress":
-        addr, data = super().deserialize(data)
+        mode, data = AddrMode.deserialize(data)
+        address, data = EUI64.deserialize(data)
 
-        if isinstance(addr.address, NWK):
-            # The address is padded
-            data = data[6:]
+        if mode != AddrMode.IEEE:
+            address, _ = NWK.deserialize(address.serialize())
 
-        return addr, data
+        return cls(mode=mode, address=address), data
 
     def serialize(self) -> bytes:
-        data = super().serialize()
+        result = (
+            self.mode.serialize() + self._get_address_type()(self.address).serialize()
+        )
 
-        if isinstance(self.address, NWK):
-            data += b"\x00\x00\x00\x00\x00\x00"
+        if self.mode != AddrMode.IEEE:
+            result += b"\x00\x00\x00\x00\x00\x00"
 
-        return data
+        return result
 
+    def __eq__(self, other):
+        if not isinstance(other, type(self)):
+            return NotImplemented
 
-class Beacon(struct.Struct):
-    """Beacon message."""
+        return self.mode == other.mode and self.address == other.address
 
-    Src: NWK
-    PanId: PanId
-    Channel: basic.uint8_t
-    PermitJoining: basic.uint8_t
-    RouterCapacity: basic.uint8_t
-    DeviceCapacity: basic.uint8_t
-    ProtocolVersion: basic.uint8_t
-    StackProfile: basic.uint8_t
-    LQI: basic.uint8_t
-    Depth: basic.uint8_t
-    UpdateId: basic.uint8_t
-    ExtendedPanId: ExtendedPanId
+    def __repr__(self) -> str:
+        return f"{type(self).__name__}(mode={self.mode!r}, address={self.address!r})"
 
 
 class GroupId(basic.uint16_t, hex_repr=True):
@@ -133,7 +158,7 @@ class Param:
     """Schema parameter"""
 
     name: str
-    type: typing.Any = None
+    type: type = None
     description: str = ""
     optional: bool = False
 
@@ -386,6 +411,8 @@ class KeySource(basic.FixedList, item_type=basic.uint8_t, length=8):
 
 
 class StartupOptions(basic.enum_flag_uint8):
+    NONE = 0
+
     ClearConfig = 1 << 0
     ClearState = 1 << 1
     AutoStart = 1 << 2
@@ -414,80 +441,3 @@ class ClusterIdList(basic.LVList, item_type=ClusterId, length_type=basic.uint8_t
 
 class NWKList(basic.LVList, item_type=NWK, length_type=basic.uint8_t):
     pass
-
-
-class TCLinkKey(struct.Struct):
-    ExtAddr: EUI64
-    Key: KeyData
-    TxFrameCounter: basic.uint32_t
-    RxFrameCounter: basic.uint32_t
-
-
-class NwkKeyDesc(struct.Struct):
-    KeySeqNum: basic.uint8_t
-    Key: KeyData
-
-
-class NwkActiveKeyItemsCC2531(struct.Struct):
-    Active: NwkKeyDesc
-    FrameCounter: basic.uint32_t
-
-
-class NwkActiveKeyItems(struct.Struct):
-    Active: NwkKeyDesc
-
-    PaddingByte1: struct.PaddingByte
-    PaddingByte2: struct.PaddingByte
-    PaddingByte3: struct.PaddingByte
-
-    FrameCounter: basic.uint32_t
-
-
-class KeyType(MissingEnumMixin, basic.enum_uint8):
-    # Standard Network Key
-    NWK = 1
-    # Application Master Key
-    APP_MASTER = 2
-    # Application Link Key
-    APP_LINK = 3
-    # Trust Center Link Key
-    TC_LINK = 4
-
-
-class KeyAttributes(basic.enum_uint8):
-    # Used for IC derived keys
-    PROVISIONAL_KEY = 0x00
-    # Unique key that is not verified
-    UNVERIFIED_KEY = 0x01
-    # Unique key that got verified by ZC
-    VERIFIED_KEY = 0x02
-
-    # Internal definitions
-
-    # Use default key to join
-    DISTRIBUTED_DEFAULT_KEY = 0xFC
-    # Joined a network which is not R21 nwk, so TCLK process finished.
-    NON_R21_NWK_JOINED = 0xFD
-    # Unique key that got verified by Joining device.
-    # This means that key is stored as plain text (not seed hashed)
-    VERIFIED_KEY_JOINING_DEV = 0xFE
-    # Entry using default key
-    DEFAULT_KEY = 0xFF
-
-
-class TCLKDevEntry(struct.Struct):
-    txFrmCntr: basic.uint32_t
-    rxFrmCntr: basic.uint32_t
-
-    extAddr: EUI64
-    keyAttributes: KeyAttributes
-    keyType: KeyType
-
-    # For Unique key this is the number of shifts
-    # for IC this is the offset on the NvId index
-    SeedShift_IcIndex: basic.uint8_t
-
-
-class NwkSecMaterialDesc(struct.Struct):
-    FrameCounter: basic.uint32_t
-    ExtendedPanID: EUI64
