@@ -1,9 +1,9 @@
+from __future__ import annotations
+
 import os
 import time
-import typing
 import asyncio
 import logging
-import warnings
 import itertools
 import contextlib
 
@@ -15,6 +15,7 @@ import zigpy.device
 import async_timeout
 import zigpy.endpoint
 import zigpy.profiles
+import zigpy.application
 import zigpy.zcl.foundation
 from zigpy.zcl import clusters
 from zigpy.types import ExtendedPanId, deserialize as list_deserialize
@@ -25,38 +26,31 @@ import zigpy_znp.types as t
 import zigpy_znp.config as conf
 import zigpy_znp.commands as c
 from zigpy_znp.api import ZNP
-from zigpy_znp.znp.nib import NIB, CC2531NIB, parse_nib
 from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
 from zigpy_znp.types.nvids import OsalNvIds
 from zigpy_znp.zigbee.zdo_converters import ZDO_CONVERTERS
 
-with warnings.catch_warnings():
-    warnings.filterwarnings(
-        action="ignore",
-        module="aiohttp",
-        message='"@coroutine" decorator is deprecated',
-        category=DeprecationWarning,
-    )
-    import zigpy.application
-
-
 ZDO_ENDPOINT = 0
 
-PROBE_TIMEOUT = 5  # seconds
-STARTUP_TIMEOUT = 5  # seconds
-ZDO_REQUEST_TIMEOUT = 15  # seconds
-DATA_CONFIRM_TIMEOUT = 8  # seconds
-DEVICE_JOIN_MAX_DELAY = 2  # seconds
-NETWORK_COMMISSIONING_TIMEOUT = 30  # seconds
-WATCHDOG_PERIOD = 30  # seconds
+# All of these are in seconds
+PROBE_TIMEOUT = 5
+STARTUP_TIMEOUT = 5
+ZDO_REQUEST_TIMEOUT = 15
+DATA_CONFIRM_TIMEOUT = 8
+DEVICE_JOIN_MAX_DELAY = 5
+NETWORK_COMMISSIONING_TIMEOUT = 30
+WATCHDOG_PERIOD = 30
+BROADCAST_SEND_WAIT_DURATION = 3
+MULTICAST_SEND_WAIT_DURATION = 3
 
 REQUEST_MAX_RETRIES = 5
-REQUEST_ERROR_RETRY_DELAY = 0.5  # second
+REQUEST_ERROR_RETRY_DELAY = 0.5  # seconds
 
 # Errors that go away on their own after waiting for a bit
 REQUEST_TRANSIENT_ERRORS = {
     t.Status.BUFFER_FULL,
     t.Status.MAC_CHANNEL_ACCESS_FAILURE,
+    t.Status.MAC_TRANSACTION_OVERFLOW,
     t.Status.MAC_NO_RESOURCES,
     t.Status.MEM_ERROR,
     t.Status.NWK_TABLE_FULL,
@@ -71,13 +65,17 @@ REQUEST_ROUTING_ERRORS = {
 
 REQUEST_RETRYABLE_ERRORS = REQUEST_TRANSIENT_ERRORS | REQUEST_ROUTING_ERRORS
 
+Z2M_PAN_ID = 0xA162
+Z2M_EXT_PAN_ID = t.EUI64.convert("DD:DD:DD:DD:DD:DD:DD:DD")
+Z2M_NETWORK_KEY = t.KeyData([1, 3, 5, 7, 9, 11, 13, 15, 0, 2, 4, 6, 8, 10, 12, 13])
+
 DEFAULT_TC_LINK_KEY = t.TCLinkKey(
     ExtAddr=t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),  # global
     Key=t.KeyData(b"ZigBeeAlliance09"),
     TxFrameCounter=0,
     RxFrameCounter=0,
 )
-ZSTACK_CONFIGURE_SUCCESS = b"\x55"
+ZSTACK_CONFIGURE_SUCCESS = t.uint8_t(0x55)
 
 LOGGER = logging.getLogger(__name__)
 
@@ -113,7 +111,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     def __init__(self, config: conf.ConfigType):
         super().__init__(config=conf.CONFIG_SCHEMA(config))
 
-        self._znp: typing.Optional[ZNP] = None
+        self._znp: ZNP | None = None
 
         # It's simpler to work with Task objects if they're never actually None
         self._reconnect_task = asyncio.Future()
@@ -122,7 +120,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         self._watchdog_task = asyncio.Future()
         self._watchdog_task.cancel()
 
-        self._nib = NIB()
         self._network_key = None
         self._network_key_seq = None
         self._concurrent_requests_semaphore = None
@@ -135,12 +132,12 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     ##################################################################
 
     @property
-    def network_key(self) -> typing.Optional[t.KeyData]:
+    def network_key(self) -> t.KeyData | None:
         # This is not a standard Zigpy property
         return self._network_key
 
     @property
-    def network_key_seq(self) -> typing.Optional[t.uint8_t]:
+    def network_key_seq(self) -> t.uint8_t | None:
         # This is not a standard Zigpy property
         return self._network_key_seq
 
@@ -188,7 +185,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             self._znp.close()
             self._znp = None
 
-    async def startup(self, auto_form=False, force_form=False):
+    async def startup(self, auto_form=False, force_form=False, read_only=False):
         """
         Performs application startup.
 
@@ -197,12 +194,16 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         """
 
         try:
-            return await self._startup(auto_form=auto_form, force_form=force_form)
+            return await self._startup(
+                auto_form=auto_form,
+                force_form=force_form,
+                read_only=read_only,
+            )
         except Exception:
             await self.shutdown()
             raise
 
-    async def _startup(self, auto_form=False, force_form=False):
+    async def _startup(self, auto_form=False, force_form=False, read_only=False):
         assert self._znp is None
 
         znp = ZNP(self.config)
@@ -222,14 +223,23 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             configured_nv_item = OsalNvIds.HAS_CONFIGURED_ZSTACK3
 
         try:
-            configured_value = await self._znp.nvram.osal_read(configured_nv_item)
+            configured_value = await self._znp.nvram.osal_read(
+                configured_nv_item, item_type=t.uint8_t
+            )
             is_configured = configured_value == ZSTACK_CONFIGURE_SUCCESS
         except KeyError:
             is_configured = False
 
-        if not is_configured or force_form:
-            if not auto_form and not force_form:
+        if force_form:
+            LOGGER.info("Forming a new network")
+            await self.form_network()
+        elif not is_configured:
+            if not auto_form:
                 raise RuntimeError("Cannot start application, network is not formed")
+            elif read_only:
+                raise RuntimeError(
+                    "Cannot start application, network is not formed and read-only"
+                )
 
             LOGGER.info("ZNP is not configured, forming a new network")
 
@@ -238,10 +248,12 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             await self.form_network()
         else:
             # Issue a reset first to make sure we aren't permitting joins
-            await self._reset()
+            await self._znp.reset()
 
             LOGGER.info("ZNP is already configured, not forming a new network")
-            await self._write_stack_settings(reset_if_changed=True)
+
+            if not read_only:
+                await self._write_stack_settings(reset_if_changed=True)
 
         # At this point the device state should the same, regardless of whether we just
         # formed a new network or are restoring one
@@ -301,12 +313,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if self.znp_config[conf.CONF_LED_MODE] is not None:
             await self._set_led_mode(led=0xFF, mode=self.znp_config[conf.CONF_LED_MODE])
 
-        device_info = await self._znp.request(
-            c.Util.GetDeviceInfo.Req(), RspStatus=t.Status.SUCCESS
-        )
-
-        self._ieee = device_info.IEEE
-        self._nwk = 0x0000
+        await self._load_network_info()
 
         # Add the coordinator as a zigpy device. We do this up here because
         # `self._register_endpoint()` adds endpoints to this device object.
@@ -338,11 +345,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             device_id=zigpy.profiles.zll.DeviceType.CONTROLLER,
         )
 
-        await self._load_device_info()
-
         # Now that we know what device we are, set the max concurrent requests
         if self.znp_config[conf.CONF_MAX_CONCURRENT_REQUESTS] == "auto":
-            max_concurrent_requests = 2 if isinstance(self._nib, CC2531NIB) else 16
+            max_concurrent_requests = 16 if self._znp.nvram.align_structs else 2
         else:
             max_concurrent_requests = self.znp_config[conf.CONF_MAX_CONCURRENT_REQUESTS]
 
@@ -355,12 +360,19 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         LOGGER.info("  Z-Stack build id: %s", version_rsp.CodeRevision)
         LOGGER.info("  Max concurrent requests: %s", max_concurrent_requests)
         LOGGER.info("  Channel: %s", self.channel)
-        LOGGER.info("  PAN ID: 0x%04x", self.pan_id)
+        LOGGER.info("  PAN ID: 0x%04X", self.pan_id)
         LOGGER.info("  Extended PAN ID: %s", self.extended_pan_id)
+        LOGGER.info("  Device IEEE: %s", self.ieee)
+        LOGGER.info("  Device NWK: 0x%04X", self.nwk)
         LOGGER.debug(
             "  Network key: %s", ":".join(f"{c:02x}" for c in self.network_key)
         )
 
+        if self.network_key == Z2M_NETWORK_KEY:
+            LOGGER.warning(
+                "Your network is using the insecure Zigbee2MQTT network key!"
+            )
+
         self._watchdog_task = asyncio.create_task(self._watchdog_loop())
 
     async def update_network_channel(self, channel: t.uint8_t):
@@ -388,7 +400,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         # The above command takes a few seconds to work
         while self.channel != channel:
-            await self._load_device_info()
+            await self._load_network_info()
             await asyncio.sleep(1)
 
     async def update_pan_id(self, pan_id: t.uint16_t) -> None:
@@ -396,9 +408,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         Updates the network PAN ID, bypassing Z-Stack anti-collision checks.
         """
 
-        await self._reset()
+        await self._znp.reset()
 
-        nib = parse_nib(await self._znp.nvram.osal_read(OsalNvIds.NIB))
+        nib = await self._znp.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
         nib.nwkPanId = pan_id
 
         await self._znp.nvram.osal_write(OsalNvIds.NIB, nib)
@@ -437,6 +449,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         # Delete any existing HAS_CONFIGURED_ZSTACK* NV items. One (or both) may fail.
         await self._znp.nvram.osal_delete(OsalNvIds.HAS_CONFIGURED_ZSTACK1)
         await self._znp.nvram.osal_delete(OsalNvIds.HAS_CONFIGURED_ZSTACK3)
+        await self._znp.nvram.osal_delete(OsalNvIds.BDBNODEISONANETWORK)
 
         # Instruct Z-Stack to reset everything on the next boot
         await self._znp.nvram.osal_write(
@@ -445,25 +458,25 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         )
 
         # And reset to clear everything
-        await self._reset()
+        await self._znp.reset()
 
         # Now that we've cleared everything, write back our Z-Stack settings
         LOGGER.debug("Updating network settings")
 
         await self._write_stack_settings(reset_if_changed=False)
 
-        # Ignore the user's PAN ID for now and just let Z-Stack pick one that works
-        await self._znp.nvram.osal_write(
-            OsalNvIds.PANID, t.uint16_t(0xFFFF), create=True
-        )
-        await self._znp.nvram.osal_write(
-            OsalNvIds.APS_USE_EXT_PANID, extended_pan_id, create=True
-        )
-        await self._znp.nvram.osal_write(OsalNvIds.PRECFGKEY, network_key, create=True)
-        await self._znp.nvram.osal_write(
-            OsalNvIds.PRECFGKEYS_ENABLE, t.Bool(True), create=True
-        )
-        await self._znp.nvram.osal_write(OsalNvIds.CHANLIST, channels, create=True)
+        for item, value in {
+            # Ignore the user's PAN ID for now and just let Z-Stack pick one that works
+            OsalNvIds.PANID: t.uint16_t(0xFFFF),
+            OsalNvIds.APS_USE_EXT_PANID: extended_pan_id,
+            # XXX: Z2M incorrectly uses this NVRAM item, APS_USE_EXT_PANID is correct
+            OsalNvIds.EXTENDED_PAN_ID: extended_pan_id,
+            OsalNvIds.PRECFGKEY: network_key,
+            # XXX: Z2M requires this item to be False
+            OsalNvIds.PRECFGKEYS_ENABLE: t.Bool(False),
+            OsalNvIds.CHANLIST: channels,
+        }.items():
+            await self._znp.nvram.osal_write(item, value, create=True)
 
         # Z-Stack Home 1.2 doesn't have the BDB subsystem
         if self._znp.version > 1.2:
@@ -497,40 +510,49 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 ),
             ]
         ):
-            if self._znp.version > 1.2:
-                # Z-Stack 3 uses the BDB subsystem
-                commissioning_rsp = await self._znp.request_callback_rsp(
-                    request=c.AppConfig.BDBStartCommissioning.Req(
-                        Mode=c.app_config.BDBCommissioningMode.NwkFormation
-                    ),
-                    RspStatus=t.Status.SUCCESS,
-                    callback=c.AppConfig.BDBCommissioningNotification.Callback(
-                        partial=True,
-                        RemainingModes=c.app_config.BDBCommissioningMode.NONE,
-                    ),
-                    timeout=NETWORK_COMMISSIONING_TIMEOUT,
-                )
+            try:
+                if self._znp.version > 1.2:
+                    # Z-Stack 3 uses the BDB subsystem
+                    commissioning_rsp = await self._znp.request_callback_rsp(
+                        request=c.AppConfig.BDBStartCommissioning.Req(
+                            Mode=c.app_config.BDBCommissioningMode.NwkFormation
+                        ),
+                        RspStatus=t.Status.SUCCESS,
+                        callback=c.AppConfig.BDBCommissioningNotification.Callback(
+                            partial=True,
+                            RemainingModes=c.app_config.BDBCommissioningMode.NONE,
+                        ),
+                        timeout=NETWORK_COMMISSIONING_TIMEOUT,
+                    )
 
-                if (
-                    commissioning_rsp.Status
-                    != c.app_config.BDBCommissioningStatus.Success
-                ):
-                    raise RuntimeError(f"Network formation failed: {commissioning_rsp}")
-            else:
-                await self._znp.nvram.osal_write(
-                    OsalNvIds.TCLK_SEED,
-                    value=DEFAULT_TC_LINK_KEY,
-                    create=True,
-                )
+                    if (
+                        commissioning_rsp.Status
+                        != c.app_config.BDBCommissioningStatus.Success
+                    ):
+                        raise RuntimeError(
+                            f"Network formation failed: {commissioning_rsp}"
+                        )
+                else:
+                    await self._znp.nvram.osal_write(
+                        OsalNvIds.TCLK_SEED,
+                        value=DEFAULT_TC_LINK_KEY,
+                        create=True,
+                    )
 
-                # In Z-Stack 1.2.2, StartupFromApp actually does what it says
-                await self._znp.request(
-                    c.ZDO.StartupFromApp.Req(StartDelay=100),
-                    RspState=c.zdo.StartupState.NewNetworkState,
-                )
+                    # In Z-Stack 1.2.2, StartupFromApp actually does what it says
+                    await self._znp.request(
+                        c.ZDO.StartupFromApp.Req(StartDelay=100),
+                        RspState=c.zdo.StartupState.NewNetworkState,
+                    )
 
-            # Both versions still end with this callback
-            await started_as_coordinator
+                # Both versions still end with this callback
+                await started_as_coordinator
+            except asyncio.TimeoutError as e:
+                raise RuntimeError(
+                    "Network formation refused, RF environment is likely too noisy."
+                    " Temporarily unscrew the antenna or shield the coordinator"
+                    " with metal until a network is formed."
+                ) from e
 
         LOGGER.debug("Waiting for the NIB to be populated")
 
@@ -539,7 +561,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         # not appear to be any user-facing MT command to read this information.
         while True:
             try:
-                nib = parse_nib(await self._znp.nvram.osal_read(OsalNvIds.NIB))
+                nib = await self._znp.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
 
                 LOGGER.debug("Current NIB is %s", nib)
 
@@ -568,7 +590,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         # Finally, reset once more to reset the device back to a "normal" state so that
         # we can continue the normal application startup sequence.
-        await self._reset()
+        await self._znp.reset()
 
     def get_dst_address(self, cluster):
         """
@@ -601,7 +623,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         data,
         expect_reply=True,
         use_ieee=False,
-    ) -> typing.Tuple[t.Status, str]:
+    ) -> tuple[t.Status, str]:
         tx_options = c.af.TransmitOptions.SUPPRESS_ROUTE_DISC_NETWORK
 
         if expect_reply:
@@ -635,7 +657,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         sequence,
         data,
         broadcast_address=zigpy.types.BroadcastAddress.RX_ON_WHEN_IDLE,
-    ) -> typing.Tuple[t.Status, str]:
+    ) -> tuple[t.Status, str]:
         assert grpid == 0
 
         return await self._send_request(
@@ -663,7 +685,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         *,
         hops=0,
         non_member_radius=3,
-    ) -> typing.Tuple[t.Status, str]:
+    ) -> tuple[t.Status, str]:
         return await self._send_request(
             dst_addr=t.AddrModeAddress(mode=t.AddrMode.Group, address=group_id),
             dst_ep=src_ep,
@@ -985,7 +1007,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         try:
             async with async_timeout.timeout(0.3):
                 await self._znp.request(
-                    c.Util.LEDControl.Req(LED=led, Mode=mode),
+                    c.UTIL.LEDControl.Req(LED=led, Mode=mode),
                     RspStatus=t.Status.SUCCESS,
                 )
         except (asyncio.TimeoutError, CommandNotRecognized):
@@ -1016,18 +1038,20 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         for nvid, value in settings.items():
             try:
-                current_value = await self._znp.nvram.osal_read(nvid)
+                current_value = await self._znp.nvram.osal_read(
+                    nvid, item_type=type(value)
+                )
             except InvalidCommandResponse:
                 current_value = None
 
             # There is no point in issuing a write if the value will not change
-            if current_value != value.serialize():
+            if current_value != value:
                 await self._znp.nvram.osal_write(nvid, value)
                 any_changed = True
 
         if reset_if_changed and any_changed:
             # Reset to make the above NVRAM writes take effect
-            await self._reset()
+            await self._znp.reset()
 
     @contextlib.asynccontextmanager
     async def _limit_concurrency(self):
@@ -1172,42 +1196,22 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             RspStatus=t.Status.SUCCESS,
         )
 
-    async def _load_device_info(self):
+    async def _load_network_info(self) -> None:
         """
         Loads low-level network information from NVRAM.
         """
 
-        # Parsing the NIB struct gives us access to low-level info, like the channel
-        self._nib = parse_nib(await self._znp.nvram.osal_read(OsalNvIds.NIB))
-        LOGGER.debug("Parsed NIB: %s", self._nib)
+        await self._znp.load_network_info()
 
-        nwkkey = await self._znp.nvram.osal_read(OsalNvIds.NWKKEY)
-
-        if self._znp.version < 3.30:
-            key_info, _ = t.NwkActiveKeyItemsCC2531.deserialize(nwkkey)
-        else:
-            key_info, _ = t.NwkActiveKeyItems.deserialize(nwkkey)
-
-        self._channel = self._nib.nwkLogicalChannel
-        self._channels = self._nib.channelList
-        self._pan_id = self._nib.nwkPanId
-        self._ext_pan_id = self._nib.extendedPANID
-
-        self._network_key = key_info.Active.Key
-        self._network_key_seq = key_info.Active.KeySeqNum
-
-        LOGGER.debug("Parsed key info: %s", key_info)
-
-    async def _reset(self) -> None:
-        """
-        Performs a soft reset within Z-Stack.
-        A hard reset resets the serial port, causing the device to disconnect.
-        """
-
-        await self._znp.request_callback_rsp(
-            request=c.SYS.ResetReq.Req(Type=t.ResetType.Soft),
-            callback=c.SYS.ResetInd.Callback(partial=True),
-        )
+        self._ieee = self._znp.network_info.ieee
+        self._nwk = self._znp.network_info.nwk
+        self._channel = self._znp.network_info.channel
+        self._channels = self._znp.network_info.channels
+        self._pan_id = self._znp.network_info.pan_id
+        self._ext_pan_id = self._znp.network_info.extended_pan_id
+        self._nwk_update_id = self._znp.network_info.nwk_update_id
+        self._network_key = self._znp.network_info.network_key
+        self._network_key_seq = self._znp.network_info.network_key_seq
 
     def _find_endpoint(self, dst_ep: int, profile: int, cluster: int) -> int:
         """
@@ -1372,17 +1376,14 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             )
 
         if dst_addr.mode == t.AddrMode.Broadcast:
-            # Broadcasts will not receive a confirmation but they still take time
-            # and use up concurrency slots
+            # Broadcasts will not receive a confirmation
             response = await self._znp.request(
                 request=request, RspStatus=t.Status.SUCCESS
             )
-
-            await asyncio.sleep(0.1 * self._nib.BroadcastDeliveryTime)
         else:
             async with async_timeout.timeout(DATA_CONFIRM_TIMEOUT):
-                # Shield from cancellation to prevent requests that time out
-                # in higher layers from missing expected responses
+                # Shield from cancellation to prevent requests that time out in higher
+                # layers from missing expected responses
                 response = await asyncio.shield(
                     self._znp.request_callback_rsp(
                         request=request,
@@ -1393,6 +1394,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                             # XXX: can this ever not match?
                             # Endpoint=src_ep,
                         ),
+                        # Multicasts eventually receive a confirmation but waiting for
+                        # it is unnecessary
+                        background=(dst_addr.mode == t.AddrMode.Group),
                     )
                 )
 
@@ -1431,7 +1435,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 ),
             )
 
-            await asyncio.sleep(0.1 * self._nib.RouteDiscoveryTime)
+            await asyncio.sleep(0.1 * 13)
         finally:
             future.set_result(True)
             del self._route_discovery_futures[nwk]
@@ -1447,18 +1451,23 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         options,
         radius,
         data,
-    ) -> typing.Tuple[t.Status, str]:
+    ) -> tuple[t.Status, str]:
         """
         Fault-tolerant wrapper around `_send_request_raw` that transparently attempts to
         repair routes and contact the device through other methods when Z-Stack errors
         are encountered.
         """
 
-        if dst_addr.mode == t.AddrMode.NWK:
-            device = self.get_device(nwk=dst_addr.address)
-        elif dst_addr.mode == t.AddrMode.IEEE:
-            device = self.get_device(ieee=dst_addr.address)
-        else:
+        try:
+            if dst_addr.mode == t.AddrMode.NWK:
+                device = self.get_device(nwk=dst_addr.address)
+            elif dst_addr.mode == t.AddrMode.IEEE:
+                device = self.get_device(ieee=dst_addr.address)
+            else:
+                device = None
+        except KeyError:
+            # Sometimes a request is sent to a device not in the database. This should
+            # work, the device object is only for recovery.
             device = None
 
         status = None
@@ -1518,12 +1527,14 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                         if status not in REQUEST_RETRYABLE_ERRORS:
                             raise
 
-                        # We cannot do anything but retry if the error is transient or
-                        # we are not sending a unicast request. Retry at least once.
+                        # Just retry if:
+                        #  - this is our first failure
+                        #  - the error is transient and is not a routing issue
+                        #  - or we are not sending a unicast request
                         if (
-                            status in REQUEST_TRANSIENT_ERRORS
-                            or attempt == 0
-                            or device is None
+                            attempt == 0
+                            or status in REQUEST_TRANSIENT_ERRORS
+                            or dst_addr.mode not in (t.AddrMode.NWK, t.AddrMode.IEEE)
                         ):
                             LOGGER.debug(
                                 "Request failed (%s), retry attempt %s of %s",
@@ -1545,15 +1556,15 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                         # to do this.
                         if (
                             status == t.Status.MAC_TRANSACTION_EXPIRED
+                            and device is not None
                             and association is None
                             and not tried_assoc_remove
                             and self._znp.version >= 3.30
                         ):
-                            # XXX: do we use NWK or IEEE?
                             association = await self._znp.request(
-                                c.Util.AssocGetWithAddress.Req(
+                                c.UTIL.AssocGetWithAddress.Req(
                                     IEEE=device.ieee,
-                                    NWK=device.nwk,
+                                    NWK=0x0000,  # IEEE takes priority
                                 )
                             )
 
@@ -1563,7 +1574,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                             ):
                                 try:
                                     await self._znp.request(
-                                        c.Util.AssocRemove.Req(IEEE=device.ieee)
+                                        c.UTIL.AssocRemove.Req(IEEE=device.ieee)
                                     )
                                     tried_assoc_remove = True
 
@@ -1574,23 +1585,26 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                                         "The UTIL.AssocRemove command is available only"
                                         " in Z-Stack 3 releases built after 20201017"
                                     )
-                        elif not tried_route_discovery:
-                            # If that doesn't work, try re-discovering the route.
-                            # While we can in theory poll and wait until it is fixed,
-                            # letting the retry mechanism deal with it simpler.
-                            await self._discover_route(device.nwk)
-                            tried_route_discovery = True
-                        elif not tried_disable_route_discovery_suppression:
-                            # Disable route discovery suppression. This appears to
-                            # generate a bit more network traffic.
-                            options &= ~c.af.TransmitOptions.SUPPRESS_ROUTE_DISC_NETWORK
-                            tried_disable_route_discovery_suppression = True
                         elif not tried_last_good_route and device is not None:
                             # `ZDO.SrcRtgInd` callbacks tell us the last path taken by
                             # messages from the device back to the coordinator. Sending
                             # packets backwards via this same route may work.
                             force_relays = (device.relays or [])[::-1]
                             tried_last_good_route = True
+                        elif (
+                            not tried_route_discovery
+                            and dst_addr.mode == t.AddrMode.NWK
+                        ):
+                            # If that doesn't work, try re-discovering the route.
+                            # While we can in theory poll and wait until it is fixed,
+                            # letting the retry mechanism deal with it simpler.
+                            await self._discover_route(dst_addr.address)
+                            tried_route_discovery = True
+                        elif not tried_disable_route_discovery_suppression:
+                            # Disable route discovery suppression. This appears to
+                            # generate a bit more network traffic.
+                            options &= ~c.af.TransmitOptions.SUPPRESS_ROUTE_DISC_NETWORK
+                            tried_disable_route_discovery_suppression = True
 
                         LOGGER.debug(
                             "Request failed (%s), retry attempt %s of %s",
@@ -1612,8 +1626,8 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             # not be able to find it again.
             if tried_assoc_remove and response is None:
                 await self._znp.request(
-                    c.Util.AssocAdd.Req(
-                        NWK=device.nwk,
+                    c.UTIL.AssocAdd.Req(
+                        NWK=association.Device.shortAddr,
                         IEEE=device.ieee,
                         NodeRelation=association.Device.nodeRelation,
                     )
